class MemoizedLazyList(list):
    """ A wrapper around a function which provides a list. Allows an instance of this class to be
        passed as if it is the list but delays evaluation of the function until list items are
        accessed or iterated. Unlike Django's `lazy`, the object appears to be a real list, so it
        can be used as `choices`, and the result is memoized for efficiency. As the result is
        memoized, this object is immutable; it should really be a tuple, but we can't set the
        items in a tuple because it's immutable, so it's a list!
    """

    def __init__(self, load_function):
        self._load_function = load_function
        self._loaded = False

    def _load(self):
        if not self._loaded:
            super().extend(list(self._load_function()))
            self._loaded = True

    def __iter__(self):
        self._load()
        return super().__iter__()

    def __getitem__(self, *args):
        self._load()
        return super().__getitem__(*args)

    def __len__(self):
        self._load()
        return super().__len__()

    def append(self, *args):
        raise TypeError("Cannot append to a MemoizedLazyList.")

    def clear(self):
        raise TypeError("Cannot clear a MemoizedLazyList.")

    def extend(self, *args):
        raise TypeError("Cannot extend a MemoizedLazyList.")

    def insert(self, *args):
        raise TypeError("Cannot insert a MemoizedLazyList.")

    def pop(self):
        raise TypeError("Cannot pop a MemoizedLazyList.")

    def remove(self, *args):
        raise TypeError("Cannot remove a MemoizedLazyList.")

    def reverse(self):
        raise TypeError("Cannot reverse a MemoizedLazyList.")

    def sort(self):
        raise TypeError("Cannot sort a MemoizedLazyList.")



from django.urls import path

from . import views


urlpatterns = [
    path("manage/", views.manage_migrations, name="massmigration_manage"),
    path("run/<str:key>/", views.run_migration, name="massmigration_run"),
    path("delete/<str:key>", views.delete_migration, name="massmigration_delete"),
]



# Standard library
import os
import re

# Third party
from django.apps.registry import apps
from django.utils.module_loading import import_string

# Mass Migration
from .constants import MIGRATIONS_FOLDER


class MigrationsStore:
    """ Stores a cache of the installed migrations to save us loading them from the filesystem
        repeatedly. Also allows us to access them by key.
    """

    def __init__(self):
        self._loaded = False
        self._all = []
        self._by_key = {}

    @property
    def all(self):
        if not self._loaded:
            self.load()
        return self._all

    @property
    def by_key(self):
        if not self._loaded:
            self.load()
        return self._by_key

    def load(self):
        """ Load all the migration instances from all migration files found in installed apps. """
        for app_config in apps.get_app_configs():
            migrations_path = os.path.join(app_config.path, MIGRATIONS_FOLDER)
            if os.path.isdir(migrations_path):
                items = os.listdir(migrations_path)
                for item in sorted(items):
                    if is_valid_migration_id(item.rstrip(".py")):
                        migration = load_migration(app_config, item)
                        self._all.append(migration)
                        self._by_key[migration.key] = migration
        self._loaded = True


store = MigrationsStore()


def is_valid_migration_name(name):
    return bool(re.match(r"^[a-z0-9_]+$", name))

def is_valid_migration_id(name):
    return bool(re.match(r"^\d{1,5}_[a-z0-9_]+$", name))


def load_migration(app_config, filename):
    """ Return in instance of the migration class from the given filename from the given app_config.
    """
    module_str = app_config.name
    migration_name = filename.rstrip(".py")
    class_path_str = f"{module_str}.{MIGRATIONS_FOLDER}.{migration_name}.Migration"
    cls = import_string(class_path_str)
    app_label = app_config.label
    instance = cls(app_label, migration_name)
    return instance



# Standard library
import logging

# Third party
from gcloudc.db import transaction
from djangae.utils import retry_on_error
from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.module_loading import import_string

# Djangae Migrations
from . import record_cache
from .exceptions import DependentMigrationNotApplied, MigrationAlreadyStarted
from .models import MigrationRecord


logger = logging.getLogger(__name__)

DEFAULT_BACKEND = "massmigration.backends.djangae.DjangaeBackend"


class BaseMigration:
    """ An operation to be performed on the database. """

    dependencies = []  # A list of (app_label, migration_name) pairs

    # This is for use for migrations which are not stored in a folder called 'migrations' inside
    # an installed app.
    app_label: str = None

    # This can be set to make a migration run on a specific backend, rather than the one that's
    # that's specified in the Django settings
    backend: str = None

    # This specifies the name of the method on the backend class which should be called to run this
    # migration. Custom migration types can be run on custom backends which support them by using
    # this attribute.
    backend_method: str = None

    def __init__(self, app_label, name):
        self.app_label = app_label
        self.name = name

    @property
    def key(self):
        """ A string which uniquely identifies this migration in the system. """
        return f"{self.app_label}:{self.name}"

    @property
    def description(self):
        # Allows the docstring to be accessed from templates despite its double underscore name
        return self.__doc__

    @property
    def backend_str(self):
        return (
            self.backend or
            getattr(settings, "MASSMIGRATION_BACKEND", None) or
            DEFAULT_BACKEND
        )

    def get_backend(self):
        backend_class = import_string(self.backend_str)
        return backend_class()

    def launch(self):
        """ Pass the migration to the backend to perform the data operation(s).
            This is what should be called by the web interface to trigger the migration.
        """
        self.check_dependencies()
        backend = self.get_backend()
        method = getattr(backend, self.backend_method)
        method(self)
        logger.info("Launched migration %s on backend %s", self.key, backend.__class__)

    def mark_as_started(self):
        """ Mark the migration as started in the database. """
        with transaction.atomic():
            _migration, created = MigrationRecord.objects.get_or_create(key=self.key)
            if not created:
                raise MigrationAlreadyStarted(
                    f"Migration {self.__class__.__name__} has already been initiated."
                )

    @retry_on_error()
    def mark_as_errored(self, error=None):
        """ Mark the migration as errored in the database. """
        # TODO: Generate a proper traceback here
        error_str = f"{error.__class__.__name__}: {error}"
        MigrationRecord.objects.filter(key=self.key).update(has_error=True, last_error=error_str)

    @retry_on_error()
    def mark_as_finished(self):
        """ Mark the migration as applied/finalized in the database. """
        with transaction.atomic():
            migration = MigrationRecord.objects.get(key=self.key)
            if migration.is_applied:
                logger.warning("Migration %s is already marked as applied.", self.key)
            else:
                migration.is_applied = True
                migration.save()
                logger.info("Migration %s finished. Marked it as applied.", self.key)

    def check_dependencies(self):
        """ Make sure that any migrations which this migration depends on have been applied. """
        # TODO: check that the specified migrations actually exist in the code.
        dependency_keys = [MigrationRecord.key_from_name_tuple(x) for x in self.dependencies]
        applied_keys = MigrationRecord.objects.filter(
            is_applied=True, key__in=dependency_keys
        ).values_list("pk", flat=True)
        for dependency_key in dependency_keys:
            if dependency_key not in applied_keys:
                raise DependentMigrationNotApplied(
                    f"Migration {self.key} depends on migration {dependency_key}, which has not "
                    "yet been applied."
                )


class SimpleMigration(BaseMigration):
    """ A migration which only needs to apply a very quick and simple change to the database which
        can be applied by one call of one function which can run in the time and memory constraints
        of one task.
    """

    backend_method = "run_simple"

    def operation(self):
        raise NotImplementedError("The `operation` method must be implemented by subclasses.")

    def wrapped_operation(self):
        logger.info("Running operation for migration %s", self.key)
        self.mark_as_started()
        try:
            self.operation()
        except Exception as error:
            self.mark_as_errored(error)
        else:
            self.mark_as_finished()


class MapperMigration(BaseMigration):
    """ A migration which calls a function on each object in a queryset. """

    backend_method = "run_mapper"

    def get_queryset(self):
        """ Returns the Django queryset which is to be mapped over. """
        raise NotImplementedError("The `get_queryset` method must be implemented by subclasses.")

    def operation(self, obj: models.Model) -> None:
        """ This is what will get called on each model instance in the queryset. """
        raise NotImplementedError("The `operation` method must be implemented by subclasses.")

    def wrapped_operation(self, obj, attempt_uuid):
        """ Call self.operation() on the object, but wrap it to catch any errors and set the
            migration as failed if necessary.
        """
        key = self.key
        record = record_cache.get_record(key)
        if record is None:
            logger.warning(
                "Migration %s no longer exists in the DB. Skipping processing operation.", key
            )
        elif record.attempt_uuid != attempt_uuid:
            logger.warning(
                "Migration %s now has attempt %s. Skipping processing operation from attempt %s.",
                key, record.attempt_uuid, attempt_uuid
            )
        elif record.has_error:
            logger.warning(
                "Migration %s is marked in the DB as having errors. Skipping processing operation.",
                key,
            )
        else:
            # We could log the object with just str(obj) here, but as the model might have a custom
            # __str__ method which does DB lookups, we just use the PK to ensure efficiency
            logger.info(
                "Running operation for migration %s on %s (pk=%r).",
                key,
                obj.__class__.__name__,
                obj.pk,
            )
            try:
                self.operation(obj)
            except Exception as error:
                logger.exception(
                    "Error in migration %s trying to process object %s (pk=%r).",
                    key,
                    obj.__class__.__name__,
                    obj.pk,
                )
                self.mark_as_errored(error)



# Standard library
import os

# Third party
from django.apps.registry import apps
from django.core.management.base import BaseCommand, CommandError
from django.template.exceptions import TemplateDoesNotExist
from django.template.loader import get_template

# Mass Migration
from massmigration.constants import MIGRATIONS_FOLDER
from massmigration.loader import is_valid_migration_id, is_valid_migration_name

class Command(BaseCommand):
    # TODO: write this

    # 1. Get name from command line args.
    # 2. Get app name from command line args.
    # 3. Get template name from command line args. Should allow either:
    #    - The name of a template in our 'templates' folder, e.g. 'mapper'; or
    #    - A path to a custom template, e.g. `myapp/migration-templates/my-migration.py`
    # 4. Make sure 'massmigrations' folder exists in that app
    # 5. Check name doesn't already exist as a file (with extension) in that folder.
    # 6. Get highest numbered migration in that folder.
    # 7. Create file with next migration number.
    # 8. Use template to populate file with empty Migration class and `dependencies` set to previous migration.

    help = "Create a new mass-migration file."

    def add_arguments(self, parser):
        parser.add_argument(
            "app_label",
            help=(
                "Name of the app for which to create a migration. "
                "This must be an installed app in your project."
            ),
        )
        parser.add_argument(
            "migration_name",
            help=(
                "Name to give the migration file. "
                "This can only contain letters, numbers and underscores."
            ),
        )
        parser.add_argument(
            "--template",
            default="simple",
            help=(
                "Which mapper template to use as the base. "
                "Must exist as a python file in massmigration/templates/migration_templates."
            ),
        )

    def handle(self, *args, **options):
        app_label = options["app_label"]
        migration_name = options["migration_name"]

        if not is_valid_migration_name(migration_name):
            raise CommandError(
                f"{migration_name} is not a valid migration name. "
                "Use letters, digits and underscores only."
            )

        try:
            template = get_template(f"migration_templates/{options['template']}.py")
        except TemplateDoesNotExist:
            raise CommandError(f"Couldn't find migration template '{options['template']}'.")

        app = apps.get_app_config(app_label)
        migrations_folder_path = os.path.join(app.path, MIGRATIONS_FOLDER)
        if not os.path.isdir(migrations_folder_path):
            os.mkdir(migrations_folder_path)

        existing_migrations = []
        for item in os.listdir(migrations_folder_path):
            if is_valid_migration_id(item.rstrip(".py")):
                existing_migrations.append(item)

        if existing_migrations:
            latest = sorted(existing_migrations, reverse=True)[0]
            highest_number = int(latest.split("_", 1)[0])
            dependencies = [(app_label, latest.rstrip(".py"))]
        else:
            highest_number = 0
            dependencies = []

        migration_number = highest_number + 1
        full_name = f"{migration_number:04d}_{migration_name}.py"

        context = {"dependencies": dependencies}
        rendered = template.render(context)
        migration_path = os.path.join(migrations_folder_path, full_name)
        with open(migration_path, "w") as file:
            file.write(rendered)

        self.stdout.write(f"Created new migration: {migration_path}")



class MigrationError(Exception):
    """ Base exception class for migration related errors. """


class MigrationAlreadyStarted(MigrationError):
    pass


class DependentMigrationNotApplied(MigrationError):
    """ Error for when trying to apply a migration which depends on another migration, and that
        other migration has not yet been applied.
    """
    pass


class RequiredMigrationNotApplied(Exception):
    """ Error for when a block of code which is marked as requiring a particular migration is being
        tried to run when that migration is not yet applied.
    """
    pass



# Third party
from django.contrib import admin

# Mass Migration
from .loader import store
from .models import MigrationRecord


class MigrationRecordAdmin(admin.ModelAdmin):
    """ Custom admin class for the Migration model. """

    ordering = ("-initiated_at",)
    list_display = ("key", "initiated_at", "is_applied", "applied_at", "was_faked")
    list_filter = ("app_label", "is_applied", "was_faked")
    readonly_fields = ("app_label", "name", "description")

    def description(self, obj):
        migration = store.by_key[obj.key]
        return migration.description


admin.site.register(MigrationRecord, MigrationRecordAdmin)



# Third party
from django.contrib import messages
from django.http import Http404
from django.shortcuts import redirect, render

# Mass Migration
from massmigration.exceptions import DependentMigrationNotApplied
from massmigration.loader import store
from massmigration.models import MigrationRecord
from massmigration.utils.permissions import superuser_required


@superuser_required()
def manage_migrations(request):
    """ A page to manage mass migrations. """
    migrations = store.all
    # Load the migration records in one query to avoid a separate query for each one
    records_by_key = MigrationRecord.objects.in_bulk()
    for migration in migrations:
        migration.record = records_by_key.get(migration.key)
    context = {
        "migrations": migrations,
    }
    return render(request, "massmigration/manage_migrations.html", context)


@superuser_required()
def run_migration(request, key):
    """ Trigger the running of a migration. """
    migration = store.by_key.get(key)
    record = MigrationRecord.objects.filter(key=key).first()
    if not migration:
        raise Http404(f"Migration with key '{key}' not found.")
    if record:
        messages.error(request, f"Migration '{key}' has already been started.")
        return redirect("massmigration_manage")
    try:
        migration.check_dependencies()
    except DependentMigrationNotApplied as error:
        messages.error(request, str(error))
        return redirect("massmigration_manage")

    if request.method == "POST":
        migration.launch()
        messages.success(request, f"Migration '{key}' started.")
        return redirect("massmigration_manage")

    # else...
    context = {
        "migration": migration,
    }
    return render(request, "massmigration/run_migration.html", context)



@superuser_required()
def delete_migration(request, key):
    """ Delete a migration which has already started or has errored. """
    migration = store.by_key.get(key)
    record = MigrationRecord.objects.filter(key=key).first()
    if not migration:
        raise Http404(f"Migration with key {key} not found.")
    if not record:
        messages.error(
            request,
            f"Can't delete migration '{key}' because no record for it exists. "
            "Maybe you deleted it already?"
        )
        return redirect("massmigration_manage")

    if request.method == "POST":
        record.delete()
        messages.success(request, f"Deleted record for migration '{key}")
        return redirect("massmigration_manage")

    # else...
    context = {
        "migration": migration,
    }
    return render(request, "massmigration/delete_migration.html", context)



from abc import ABC


class BackendBase(ABC):
    """ Abstract base class which defines the methods that a backend must implement. """

    def run_simple(self, migration):
        """ Run the wrapped_operation() method on the given SimpleMigration.
            For simple migrations, the wrapped_operation() method handles the calling of
            mark_as_started(), mark_as_errored() and mark_as_finished() for you.
            The backend just needs to hand off to whatever processor/task service it wants to in
            order to call wrapped_operation().
        """
        raise NotImplementedError

    def run_mapper(self, migration):
        """ Call migration.wrapped_operation(instance, attempt_uuid) for every instance returned by
            the get_queryset() method of the given MapperMigration.
            This MUST:
            * Call migration.mark_as_started() before performing the data operation.
            * Get the value of `migration.attempt_uuid` when this method was called, and pass it as
              the second argument to every call of `wrapped_operation()`.
            * Call migration.mark_as_finished() when the end of the queryset is reached.

            If `wrapped_operation()` returns False, it means that the migration has errored and
            iteration can be stopped. But continuing to iterate will do no harm.
        """
        raise NotImplementedError



""" Utilities for getting information about MigrationRecord objects without hammering the DB too much."""

# Third party
from django.conf import settings
from django.core.cache import cache
from django.db.models.signals import post_save

# Mass Migration
from massmigration.models import MigrationRecord


DEFAULT_CACHE_TIMEOUT = 60


def get_record(key):
    """ Get the MigrationRecord for the given key. This is designed to be called heavily, i.e. for
        every object in a MapperMigration, so the result is cached with a best-effort attempt to
        refresh that cache when the record changes, but with no guarantee of it.
    """
    cache_key = get_cache_key(key)
    record = cache.get(cache_key)
    if not record:
        record = MigrationRecord.objects.filter(key=key).first()
        cache.set(cache_key, record, cache_timeout())
    return record


def get_cache_key(migration_key):
    return f"massmigration_record:{migration_key}"


def cache_timeout():
    return getattr(settings, "MASSMIGRATION_RECORD_CACHE_TIMEOUT", DEFAULT_CACHE_TIMEOUT)


def record_post_save(sender, **kwargs):
    """ Update the cache when a MigrationRecord is changed (the relevant scenarios being when it's
        marked as started, marked as errored or marked as finished).
    """
    record = kwargs["instance"]
    cache_key = get_cache_key(record.key)
    cache.set(cache_key, record, cache_timeout())


post_save.connect(record_post_save, sender=MigrationRecord)



from django.db import models
from gcloudc.db.models.fields.computed import ComputedFieldMixin


# TODO: make a MR to put this into gcloudc
class ComputedDateTimeField(ComputedFieldMixin, models.DateTimeField):
    pass


# TODO: make a MR to make this the defualt ComputedCharField in gcloudc
class ComputedCharField(ComputedFieldMixin, models.CharField):
    pass



from django.contrib.auth.decorators import user_passes_test


def superuser_required(login_url=None):
    """ Decorator for views that checks whether a user is a superuser. """

    def is_superuser(user):
        return user.is_superuser

    return user_passes_test(is_superuser, login_url=login_url)



# Standard library
import uuid

# Third party
from django.db import models
from django.utils import timezone
from gcloudc.db.models.fields.computed import ComputedBooleanField

# Djangae Migrations
from .fields import ComputedCharField, ComputedDateTimeField
from .utils.apps import get_app_label_choices
from .utils.functional import MemoizedLazyList


class MigrationRecord(models.Model):
    """ Stores a record of a particular migration being applied to the database. """

    # I'm still not sure whether the key should be computed from the app_label and name or the
    # other way round. The app_label at least is good for filtering in the Django admin though.
    key = models.CharField(max_length=250, primary_key=True)
    app_label = ComputedCharField(
        "_app_label", max_length=100, choices=MemoizedLazyList(get_app_label_choices)
    )
    name = ComputedCharField("_name", max_length=150)
    attempt_uuid = models.UUIDField(
        default=uuid.uuid4,
        help_text=(
            "A unique ID which allows us to detect if this object was deleted and recreated (e.g. "
            "due to an error), and therefore allows us to abort any stale tasks which were spawned "
            "as part of the previous attempt of the same migration."
        ),
        editable=False,
    )
    # We don't have an `is_initiated` field, as if it's not initiated it won't exist in the DB.
    initiated_at = models.DateTimeField(default=timezone.now, editable=False)
    in_progress = ComputedBooleanField("_in_progress")
    is_applied = models.BooleanField(
        default=False, help_text="Is the migration fully applied to the DB?"
    )
    applied_at = ComputedDateTimeField("_applied_at", null=True)
    has_error = models.BooleanField(default=False)
    last_error = models.TextField(blank=True)
    was_faked = models.BooleanField(default=False)

    def _app_label(self):
        return self.key.split(":")[0]

    def _name(self):
        return self.key.split(":")[1]

    def _applied_at(self):
        if self.is_applied and not self.applied_at:
            return timezone.now()
        return self.applied_at

    def _in_progress(self):
        return self.initiated_at and not self.applied_at

    @staticmethod
    def key_from_name_tuple(name_tuple):
        """ Given a 2-part migration identifier, return the single string object which we use as
            the `key` field value.
        """
        assert len(name_tuple) == 2
        return ":".join(name_tuple)



"""
Utilities for allowing you to prevent particular code functions from being run unless specific
migrations have been applied.
"""

# Standard library
from functools import wraps
import logging

# Third party
from django.http import HttpResponse

# Djangae Migrations
from .exceptions import RequiredMigrationNotApplied
from .loader import is_valid_migration_id
from .models import MigrationRecord

logger = logging.getLogger(__name__)

APPLIED_MIGRATIONS_CACHE = {}


def migration_is_applied(migration_identifier):
    """ Tells you whether or not the specified migration has been applied to the DB.
        Positive (True) responses are cached to avoid repeated DB queries.
    """
    if is_valid_migration_id(migration_identifier):
        key = migration_identifier
    else:
        key = MigrationRecord.key_from_name_tuple(migration_identifier)
    try:
        return APPLIED_MIGRATIONS_CACHE[key]
    except KeyError:
        applied = MigrationRecord.objects.filter(key=key, is_applied=True).exists()
        if applied:
            APPLIED_MIGRATIONS_CACHE[key] = True
            return True
    return False


def requires_migration(migration_identifier):
    """ Function decorator which prevents the function being run if the specified migration is not
        applied.
    """
    def decorator(function):
        @wraps(function)
        def replacement(*args, **kwargs):
            if not migration_is_applied(migration_identifier):
                raise RequiredMigrationNotApplied(
                    f"Function '{function}'' requires migration {migration_identifier} which has "
                    "not been applied."
                )
            return function(*args, **kwargs)
        return replacement
    return decorator


def view_requires_migration(migration_identifier):
    """ Same as `requires_migration`, but for view functions. Returns a 503 status HttpResponse
        rather than raising an exception.
    """
    def decorator(function):
        @wraps(function)
        def replacement(*args, **kwargs):
            if not migration_is_applied(migration_identifier):
                logger.error(
                    "View function '%s' requires migration '%s' which has not yet been applied.",
                    function, migration_identifier
                )
                return HttpResponse(
                    "This resource requires data changes which have not yet been made.",
                    status=503,
                )
            return function(*args, **kwargs)
        return replacement
    return decorator



# Standard library
import logging

# Third party
from djangae.tasks.deferred import defer, defer_iteration_with_finalize
from django.conf import settings

# Massmigration
from .base import BackendBase

logger = logging.getLogger(__name__)


class DjangaeBackend(BackendBase):
    """ Djangae-based backend for running operations on Google Cloud Tasks. """

    def run_simple(self, migration):
        defer(migration.wrapped_operation, _queue=self._get_queue_name(migration))
        logger.info("Deferred task to run single-task migration %s", migration.key)

    def run_mapper(self, migration):
        attempt_uuid = migration.attempt_uuid
        migration.mark_as_started()
        defer_iteration_with_finalize(
            migration.get_queryset(),
            self._call_mapper_wrapped_operation,
            self._mark_mapper_as_finished,
            migration=migration,
            attempt_uuid=attempt_uuid,
            _queue=self._get_queue_name(migration),
        )
        logger.info("Deferred task to run mapper migration %s", migration.key)

    def _get_queue_name(self, migration):
        """ Get the queue name from settings, or the override on the migration, if set."""
        queue = getattr(migration, "queue_name", None)
        if not queue:
            queue = getattr(settings, "MASSMIGRATION_TASK_QUEUE", None)
        if not queue:
            raise NotImplementedError("Please configure settings.MASSMIGRATION_TASK_QUEUE.")
        return queue

    def _call_mapper_wrapped_operation(self, instance, migration, attempt_uuid):
        migration.wrapped_operation(instance, attempt_uuid)

    def _mark_mapper_as_finished(self, migration, attempt_uuid):
        migration.mark_as_finished()



from setuptools import setup, find_packages
import os

NAME = "django-mass-migration"
PACKAGES = find_packages()
DESCRIPTION = "Django app for long-running data migrations"
LONG_DESCRIPTION = open(os.path.join(os.path.dirname(__file__), "README.md")).read()
URL = "https://github.com/adamalton/django-mass-migration"
AUTHOR = "Adam Alton"

# TODO: possibly make the dependency on djangae optional
EXTRAS = {}


setup(
    name=NAME,
    version="v0.1.0",
    packages=PACKAGES,
    author=AUTHOR,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    keywords=["django", "djangae", "django-gcloud-connectors", "Google App Engine"],
    url=URL,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.1",
        "Framework :: Django :: 4.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    include_package_data=True,
    install_requires=[
        "django>=3.2,<5.0",
        "djangae>=1.1.0",
        "django-gcloud-connectors>=0.3.7",
    ],
    extras_require=EXTRAS,
    # tests_require=EXTRAS["test"],
)
